@@ -10,6 +10,7 @@ module Database.Selda.PostgreSQL
 import Data.Monoid
 #endif
 import Data.ByteString (ByteString)
+import Data.String (IsString (..))
 import qualified Data.Text as T
 import Database.Selda.Backend hiding (toText)
 import Database.Selda.JSON
@@ -23,6 +24,7 @@ import qualified Data.ByteString as BS (foldl')
 import qualified Data.ByteString.Char8 as BS (pack, unpack)
 import Data.Dynamic
 import Data.Foldable (for_)
+import Data.Int (Int64)
 import Data.Text.Encoding
 import Database.Selda.PostgreSQL.Encoding
 import Database.PostgreSQL.LibPQ hiding (user, pass, db, host)
@@ -49,6 +51,17 @@ data PGConnectInfo = PGConnectInfo
     -- | Password for authentication, if necessary.
   , pgPassword :: Maybe T.Text
   }
+  | PGConnectionString
+  { -- | Custom connection PostgreSQL connection string.
+    pgConnectionString :: T.Text
+  , pgSchema :: Maybe T.Text
+  }
+
+instance IsString PGConnectInfo where
+  fromString s = PGConnectionString
+    { pgConnectionString = fromString s
+    , pgSchema = Nothing
+    }
 
 -- | Connect to the given database on the given host, on the default PostgreSQL
 --   port (5432):
@@ -98,6 +111,7 @@ pgConnString PGConnectInfo{..} = mconcat
   , "connect_timeout=10", " "
   , "client_encoding=UTF8"
   ]
+pgConnString PGConnectionString{..} = encodeUtf8 pgConnectionString
 #endif
 
 -- | Perform the given computation over a PostgreSQL database.
@@ -178,9 +192,9 @@ pgPPConfig = defPPConfig
     bigserialQue :: [ColAttr]
     bigserialQue = [AutoPrimary Strong, Required]
 
-    -- For when we use 'autoPrimaryGen' on 'Int' field
+    -- For when we use 'autoPrimary' on 'Int' field
     isGenericIntPrimaryKey :: SqlTypeRep -> [ColAttr] -> Bool
-    isGenericIntPrimaryKey ty attrs = ty == TInt && and ((`elem` attrs) <$> bigserialQue)
+    isGenericIntPrimaryKey ty attrs = ty == TInt64 && and ((`elem` attrs) <$> bigserialQue)
 
 -- | Create a `SeldaBackend` for PostgreSQL `Connection`
 pgBackend :: Connection   -- ^ PostgreSQL connection object.
@@ -252,7 +266,7 @@ pgGetTableInfo c tbl = do
     Right (_, vals) <- pgQueryRunner c False tableinfo []
     if null vals
       then do
-        pure $ TableInfo [] [] []
+        pure $ TableInfo (mkTableName tbl) [] [] []
       else do
         Right (_, pkInfo) <- pgQueryRunner c False pkquery []
         Right (_, us) <- pgQueryRunner c False uniquequery []
@@ -260,12 +274,12 @@ pgGetTableInfo c tbl = do
         Right (_, fks) <- pgQueryRunner c False fkquery []
         Right (_, ixs) <- pgQueryRunner c False ixquery []
         colInfos <- mapM (describe fks (map toText ixs)) vals
-        x <- pure $ TableInfo
-          { tableColumnInfos = colInfos
+        pure $ TableInfo
+          { tableInfoName = mkTableName tbl
+          , tableColumnInfos = colInfos
           , tableUniqueGroups = map (map mkColName) uniques
           , tablePrimaryKey = [mkColName pk | [SqlString pk] <- pkInfo]
           }
-        pure x
   where
     splitNames = breakNames . toText
     -- TODO: this is super ugly; should really be fixed
@@ -279,7 +293,9 @@ pgGetTableInfo c tbl = do
     tableinfo = mconcat
       [ "SELECT column_name, data_type, is_nullable, column_default LIKE 'nextval(%' "
       , "FROM information_schema.columns "
-      , "WHERE table_name = '", tbl, "';"
+      , "WHERE table_name = '", tbl, "' "
+      , "AND table_schema = current_schema() "
+      , "ORDER BY ordinal_position;"
       ]
     pkquery = mconcat
       [ "SELECT a.attname "
@@ -306,7 +322,7 @@ pgGetTableInfo c tbl = do
       , "  ON tc.constraint_name = kcu.constraint_name "
       , "JOIN information_schema.constraint_column_usage AS ccu "
       , "  ON ccu.constraint_name = tc.constraint_name "
-      , "WHERE constraint_type = 'FOREIGN KEY' AND tc.table_name='", tbl, "';"
+      , "WHERE constraint_type = 'FOREIGN KEY' AND tc.table_name='", tbl, "' AND tc.table_schema=current_schema();"
       ]
     ixquery = mconcat
       [ "select a.attname as column_name "
@@ -320,7 +336,7 @@ pgGetTableInfo c tbl = do
       , "and not ix.indisunique "
       , "and not ix.indisprimary "
       , "and t.relkind = 'r' "
-      , "and t.relname = '", tbl , "';"
+      , "and t.oid = '\"", tbl , "\"'::regclass;"
       ]
     describe fks ixs [SqlString name, SqlString ty, SqlString nullable, auto] =
       return $ ColumnInfo
@@ -342,7 +358,7 @@ pgGetTableInfo c tbl = do
     describe _ _ results =
       throwM $ SqlError $ "bad result from table info query: " ++ show results
 
-pgQueryRunner :: Connection -> Bool -> T.Text -> [Param] -> IO (Either Int (Int, [[SqlValue]]))
+pgQueryRunner :: Connection -> Bool -> T.Text -> [Param] -> IO (Either Int64 (Int, [[SqlValue]]))
 pgQueryRunner c return_lastid q ps = do
     mres <- execParams c (encodeUtf8 q') [fromSqlValue p | Param p <- ps] Binary
     unlessError c errmsg mres $ \res -> do
@@ -354,7 +370,7 @@ pgQueryRunner c return_lastid q ps = do
     q' | return_lastid = q <> " RETURNING LASTVAL();"
        | otherwise     = q
 
-    getLastId res = (maybe 0 id . fmap readInt) <$> getvalue res 0 0
+    getLastId res = (maybe 0 id . fmap readInt64) <$> getvalue res 0 0
 
 pgRun :: Connection -> Dynamic -> [Param] -> IO (Int, [[SqlValue]])
 pgRun c hdl ps = do
@@ -427,8 +443,9 @@ doError c msg = do
 
 mkTypeRep :: T.Text ->  Either T.Text SqlTypeRep
 mkTypeRep "bigserial"                = Right TRowID
-mkTypeRep "int8"                     = Right TInt
-mkTypeRep "bigint"                   = Right TInt
+mkTypeRep "int4"                     = Right TInt32
+mkTypeRep "int8"                     = Right TInt64
+mkTypeRep "bigint"                   = Right TInt64
 mkTypeRep "float8"                   = Right TFloat
 mkTypeRep "double precision"         = Right TFloat
 mkTypeRep "timestamp with time zone" = Right TDateTime
@@ -444,7 +461,8 @@ mkTypeRep typ                        = Left typ
 -- | Custom column types for postgres.
 pgColType :: PPConfig -> SqlTypeRep -> T.Text
 pgColType _ TRowID    = "BIGINT"
-pgColType _ TInt      = "INT8"
+pgColType _ TInt64    = "INT8"
+pgColType _ TInt32    = "INT4"
 pgColType _ TFloat    = "FLOAT8"
 pgColType _ TDateTime = "TIMESTAMP"
 pgColType _ TBlob     = "BYTEA"
